from __future__ import annotations

from enum import StrEnum


class RunMode(StrEnum):
    """Execution mode propagated from the engine to lower-level components."""

    INFERENCE = "inference"
    CONFLICT_CHECK = "conflict_check"
